/*
  to build (with OpenWatcom)
  
  wcl -zq -s conchk.c
*/

#include <ctype.h>
#include <dos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* needed for recent openwatcom... (or specify option -zp1) */
#pragma pack(1)

struct devhdr {
    void far *      next_dev;
    unsigned short  devattr;
    unsigned short  strategy;
    unsigned short  commands;
    char            name[8];
};

struct reqhdr {
    unsigned char   length;
    unsigned char   unit;
    unsigned char   command;
    unsigned short  status;
    unsigned char   reserved5[8];
};

struct reqhdr_common {
    struct reqhdr   rh;
};

struct reqhdr_inout {
    struct reqhdr   rh;
    unsigned char   media_descriptor;
    void far *      buffer;
    unsigned short  count;
    unsigned short  start_sector;
    void far *      volume_id;
    unsigned long   start_sector32;
};

struct reqhdr_nondesc {
    struct reqhdr           rh;
    unsigned char           nextinput;
};





unsigned short  call_device(const struct devhdr far *device, void far *reqh)
{
    void far *dev_strategy;
    void far *dev_commands;
    unsigned short devseg;
    
    devseg = FP_SEG(device);
    dev_strategy = MK_FP(devseg, device->strategy);
    dev_commands = MK_FP(devseg, device->commands);
    
    _asm {
        pushf
        push ax
        push bx
        push cx
        push dx
        push si
        push di
        push bp
        push ds
        push es
        les bx, dword ptr [reqh]
        call dword ptr [dev_strategy]
        call dword ptr [dev_commands]
        pop es
        pop ds
        pop bp
        pop di
        pop si
        pop dx
        pop cx
        pop bx
        pop ax
        popf
    }
    
    return ((struct reqhdr far *)reqh)->status;
}


unsigned get_version(unsigned *major, unsigned *minor, int os2flag)
{
    union REGS r;
    unsigned mj, mn;
    
    r.x.ax = 0x3000;
    intdos(&r, &r);
    mj = r.h.al;
    mn = r.h.ah;
    r.x.ax = 0x3306;
    r.x.bx = 0xffff;
    intdos(&r, &r);
    if (r.h.al != 0xff && r.h.bh < 100 && r.h.bl < 100) {
        mj = r.h.bl;
        mn = r.h.bh;
    }
    if (mj == 0) mj = 1;
    if (os2flag) {
        if (mj == 10) { mj = 3; mn = 10; /* OS/2 1.x          ... DOS 3.1 */ }
        if (mj == 20) { mj = 5; mn = 0;  /* OS/2 2.x and warp ... DOS 5.0 */ }
    }
    
    if (major) *major = mj;
    if (minor) *minor = mn;
    return (mj << 8) | mn;
}

struct devhdr far * get_nuldev(void)
{
    union REGS r;
    struct SREGS sr;
    unsigned off;
    unsigned major, minor;
    
    off = 0x22;
    get_version(&major, &minor, 1);
    if (major <= 1) return NULL;
    if (major == 2) off = 0x17;
    if (major == 3 && minor < 10) off = 0x28;
    
    r.h.ah = 0x52;
    intdosx(&r, &r, &sr);
    return MK_FP(sr.es, r.x.bx + off);
}


struct devhdr far * find_chardev(char *dev_name)
{
    struct devhdr far *rh;
    char n[8];
    int i;
    
    memset(n, ' ', 8);
    for(i=0; i<8; ++i) {
        if (!dev_name[i]) break;
        n[i] = toupper(dev_name[i]);
    }
    
    rh = get_nuldev();
    if (!rh) return NULL;
    do {
        if (_fmemcmp(rh->name, n, 8) == 0) {
            return rh;
        }
    } while ( (rh = rh->next_dev) != MK_FP(0xffffU, 0xffffU));
    
    return NULL;
}



unsigned short
dev_getnodesc(const struct devhdr far *dh, unsigned char *c)
{
    unsigned short rc;
    struct reqhdr_nondesc rh;
    
    memset(&rh, 0, sizeof(rh));
    rh.rh.length = sizeof(rh);
    rh.rh.command = 5;
    rc = call_device(dh, &rh);
    
    if (c) {
        *c = (rc & 0x200) ? 0 : rh.nextinput;
    }
    
    return rc;
}

unsigned short
dev_inputstatus(const struct devhdr far *dh)
{
    struct reqhdr_nondesc rh;
    
    memset(&rh, 0, sizeof(rh));
    rh.rh.length = sizeof(rh);
    rh.rh.command = 6;
    return call_device(dh, &rh);
}

unsigned short
dev_getc(const struct devhdr far *dh, int *c)
{
    unsigned short rc;
    struct reqhdr_inout rh;
    unsigned char buffer[1];
    
    memset(&rh, 0, sizeof(rh));
    rh.rh.length = sizeof(rh);
    rh.rh.command = 4;
    rh.count = 1;
    rh.buffer = buffer;
    rc = call_device(dh, &rh);
    if (c) {
        if (rh.count == 0) *c = EOF;
        else *c = (rc & 0x8000) ? 0 : buffer[0];
    }
    return rc;
}

void
do_idle(void)
{
    _asm {
        pushf
        sti
        nop
        hlt
        popf
    }
}


int main(int argc, char *argv[])
{
    struct devhdr far *  dh;
    unsigned short stat_is, stat_isb, stat_nd, stat_ndb;
    unsigned char c_nd;
    
    dh = get_nuldev();
    printf("NUL %Fp\n", dh);
    dh = find_chardev("CON");
    printf("CON %Fp\n", dh);
    
    stat_isb = stat_ndb = -1;
    
    printf("press 'q' key to exit...\n");
    for(;;) {
        int do_update;
        stat_nd = dev_getnodesc(dh, &c_nd);
        stat_is = dev_inputstatus(dh);
        do_update = stat_is != stat_isb || stat_nd != stat_ndb;
        if (do_update) {
            unsigned n0, n1, ptr0, ptr1;
            n0 = *(unsigned char far *)0x00600103UL;
            ptr0 = *(unsigned short far *)0x00600104UL;
            printf("input status=0x%04x  non destructive read=0x%04x (char=0x%02x)\n", stat_is, stat_nd, c_nd);
            if ((stat_nd & 0x8300) == 0x0100) {
                int c;
                dev_getc(dh, &c);
                n1 = *(unsigned char far *)0x00600103UL;
                ptr1 = *(unsigned short far *)0x00600104UL;
                printf("  input char=0x%02x", c);
                if (c >= ' ' && c < 0x7f) printf(" '%c'", c);
                printf("  buffer %04X %u -> %04X %u\n", ptr0, n0, ptr1, n1);
                if (c == 'q') break;
                if (c == 'n') fprintf(stderr, "\x1b" "[6n");
            }
            else
              printf("buffer %04X %u\n", ptr0, n0);
            stat_isb = stat_is;
            stat_ndb = stat_nd;
        } else {
            do_idle();
        }
    }
    
    return 0;
}

